"""SMARTS representations custom reactions."""
from synthemol.reactions.query_mol import QueryMol
from synthemol.reactions.reaction import Reaction

# To use custom chemical reactions instead of Enamine REAL reactions, replace None with a list of Reaction objects.
# If CUSTOM_REACTIONS is None, synthemol will default to the reactions in real.py.
CUSTOM_REACTIONS: tuple[Reaction] | None = None



"""Contains constants shared throughout synthemol."""
from importlib import resources
from typing import Literal

from rdkit.Chem import Mol
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor


CHEMBL_SMILES_COL = 'Smiles'
REAL_SPACE_SIZE = 31507987117  # As of August 30, 2022, in the 2022 q1-2 REAL space
REAL_REACTION_COL = 'reaction'
REAL_BUILDING_BLOCK_COLS = ['reagent1', 'reagent2', 'reagent3', 'reagent4']
REAL_BUILDING_BLOCK_ID_COL = 'reagent_id'
REAL_TYPE_COL = 'Type'
REAL_SMILES_COL = 'smiles'
SMILES_COL = 'smiles'
ACTIVITY_COL = 'activity'
SCORE_COL = 'score'
MODEL_TYPE = Literal['random_forest', 'mlp', 'chemprop']
FINGERPRINT_TYPES = Literal['morgan', 'rdkit']
MOLECULE_TYPE = str | Mol  # Either a SMILES string or an RDKit Mol object
SKLEARN_MODEL_TYPES = RandomForestClassifier | RandomForestRegressor | MLPClassifier | MLPRegressor
SKLEARN_MODEL_NAME_TYPES = Literal['random_forest', 'mlp']
MODEL_TYPES = Literal['random_forest', 'mlp', 'chemprop']
DATASET_TYPES = Literal['classification', 'regression']
OPTIMIZATION_TYPES = Literal['maximize', 'minimize']

# Path where data files are stored
with resources.path('synthemol', 'resources') as resources_dir:
    DATA_DIR = resources_dir

# If using custom building blocks, replace BUILDING_BLOCKS_PATH and set REACTION_TO_BUILDING_BLOCKS_PATH to None
BUILDING_BLOCKS_PATH = DATA_DIR / 'real' / 'building_blocks.csv'
REACTION_TO_BUILDING_BLOCKS_PATH = DATA_DIR / 'real' / 'reaction_to_building_blocks.pkl'



import itertools
from collections import Counter
from functools import partial
from typing import Callable

import numpy as np
from rdkit import Chem
from tqdm import trange

from synthemol.constants import OPTIMIZATION_TYPES
from synthemol.generate.node import Node
from synthemol.reactions import Reaction
from synthemol.utils import random_choice


class Generator:
    """A class that generates molecules."""

    def __init__(
            self,
            building_block_smiles_to_id: dict[str, int],
            max_reactions: int,
            scoring_fn: Callable[[str], float],
            explore_weight: float,
            num_expand_nodes: int | None,
            optimization: OPTIMIZATION_TYPES,
            reactions: tuple[Reaction],
            rng_seed: int,
            no_building_block_diversity: bool,
            store_nodes: bool,
            verbose: bool,
            replicate: bool = False
    ) -> None:
        """Creates the Generator.

        :param building_block_smiles_to_id: A dictionary mapping building block SMILES to their IDs.
        :param max_reactions: The maximum number of reactions to use to construct a molecule.
        :param scoring_fn: A function that takes as input a SMILES representing a molecule and returns a score.
        :param explore_weight: The hyperparameter that encourages exploration.
        :param num_expand_nodes: The number of tree nodes to expand when extending the child nodes in the search tree.
                                  If None, then all nodes are expanded.
        :param optimization: Whether to maximize or minimize the score.
        :param reactions: A tuple of reactions that combine molecular building blocks.
        :param rng_seed: Seed for the random number generator.
        :param no_building_block_diversity: Whether to turn off the score modification that encourages diverse building blocks.
        :param store_nodes: Whether to store the child nodes of each node in the search tree.
                            This doubles the speed of the search but significantly increases
                            the memory usage (e.g., 450GB for 20,000 rollouts instead of 600 MB).
        :param verbose: Whether to print out additional statements during generation.
        :param replicate: This is necessary to replicate the results from the paper, but otherwise should not be used
                          since it limits the potential choices of building blocks.
        """
        self.building_block_smiles_to_id = building_block_smiles_to_id
        self.max_reactions = max_reactions
        self.scoring_fn = scoring_fn
        self.explore_weight = explore_weight
        self.num_expand_nodes = num_expand_nodes
        self.optimization = optimization
        self.reactions = reactions
        self.rng = np.random.default_rng(seed=rng_seed)
        self.building_block_diversity = not no_building_block_diversity
        self.store_nodes = store_nodes
        self.verbose = verbose

        # Get all building blocks that are used in at least one reaction
        if replicate:
            self.all_building_blocks = list(dict.fromkeys(
                building_block
                for reaction in self.reactions
                for reactant in reaction.reactants
                for building_block in reactant.allowed_building_blocks
            ))
        else:
            self.all_building_blocks = sorted(
                building_block
                for building_block in self.building_block_smiles_to_id
                if any(reactant.has_match(building_block) for reaction in reactions for reactant in reaction.reactants)
            )

        # Get the function to use for optimization
        if self.optimization == 'maximize':
            self.optimization_fn = max
            self.ascending_scores = False
        elif self.optimization == 'minimize':
            self.optimization_fn = min
            self.ascending_scores = True
        else:
            raise ValueError(f'Invalid optimization type: {self.optimization}')

        # Initialize the root node
        self.rollout_num = 0
        self.root = Node(
            explore_weight=explore_weight,
            scoring_fn=scoring_fn,
            node_id=0,
            rollout_num=self.rollout_num
        )

        # Initialize the rollout num, node map, building block counts, and node to children
        self.node_map: dict[Node, Node] = {self.root: self.root}
        self.building_block_counts = Counter()
        self.node_to_children: dict[Node, list[Node]] = {}

    def get_next_building_blocks(self, molecules: tuple[str]) -> list[str]:
        """Get the next building blocks that can be added to the given molecules.

        :param molecules: A tuple of SMILES strings representing the molecules to get the next building blocks for.
        :return: A list of SMILES strings representing the next building blocks that can be added to the given molecules.
        """
        # Initialize list of allowed building blocks
        available_building_blocks = []

        # Loop through each reaction
        for reaction in self.reactions:
            # Get indices of the reactants in this reaction
            reactant_indices = set(range(reaction.num_reactants))

            # Skip reaction if there's no room to add more reactants
            if len(molecules) >= reaction.num_reactants:
                continue

            # For each molecule, get a list of indices of reactants it matches
            reactant_matches_per_molecule = [
                reaction.get_reactant_matches(smiles=molecule)
                for molecule in molecules
            ]

            # Loop through products of reactant indices that the molecules match to
            # and for each product, if it matches to all separate reactants,
            # then include the missing reactants in the set of unfilled reactants
            for matched_reactant_indices in itertools.product(*reactant_matches_per_molecule):
                matched_reactant_indices = set(matched_reactant_indices)

                if len(matched_reactant_indices) == len(molecules):
                    for index in sorted(reactant_indices - matched_reactant_indices):
                        available_building_blocks += reaction.reactants[index].allowed_building_blocks

        # Remove duplicates but maintain order for reproducibility
        available_building_blocks = list(dict.fromkeys(available_building_blocks))

        return available_building_blocks

    def get_reactions_for_molecules(self, molecules: tuple[str]) -> list[tuple[Reaction, dict[str, int]]]:
        """Get all reactions that can be run on the given molecules.

        :param molecules: A tuple of SMILES strings representing the molecules to run reactions on.
        :return: A list of tuples, where each tuple contains a reaction and a dictionary mapping
                 the molecules to the indices of the reactants they match.
        """
        matching_reactions = []

        # Check each reaction to see if it can be run on the given molecules
        for reaction in self.reactions:
            # Skip reaction if the number of molecules doesn't match the number of reactants
            if len(molecules) != reaction.num_reactants:
                continue

            # For each molecule, get a list of indices of reactants it matches
            reactant_matches_per_molecule = [
                reaction.get_reactant_matches(smiles=molecule)
                for molecule in molecules
            ]

            # Include every assignment of molecules to reactants that fills all the reactants
            for matched_reactant_indices in itertools.product(*reactant_matches_per_molecule):
                if len(set(matched_reactant_indices)) == reaction.num_reactants:
                    molecule_to_reactant_index = dict(zip(molecules, matched_reactant_indices))
                    matching_reactions.append((reaction, molecule_to_reactant_index))

        return matching_reactions

    def run_all_reactions(self, node: Node) -> list[Node]:
        """Run all possible reactions for the molecules in the Node and return the resulting product Nodes.

        :param node: A Node to run reactions for.
        :return: A list of Nodes for the products of the reactions.
        """
        # Get all reactions that are possible for the molecules in the Node
        matching_reactions = self.get_reactions_for_molecules(molecules=node.molecules)

        # Run all possible reactions and create Nodes for the products
        product_nodes = []
        product_set = set()
        for reaction, molecule_to_reactant_index in matching_reactions:
            # Put molecules in the right order for the reaction
            molecules = sorted(node.molecules, key=lambda frag: molecule_to_reactant_index[frag])

            # Run reaction
            products = reaction.run_reactants(molecules)

            if len(products) == 0:
                raise ValueError('Reaction failed to produce products.')

            assert all(len(product) == 1 for product in products)

            # Convert product mols to SMILES (and remove Hs)
            products = [Chem.MolToSmiles(Chem.RemoveHs(product[0])) for product in products]

            # Filter out products that have already been created and deduplicate
            products = list(dict.fromkeys(product for product in products if product not in product_set))
            product_set |= set(products)

            # Create reaction log
            reaction_log = {
                'reaction_id': reaction.id,
                'building_block_ids': tuple(
                    self.building_block_smiles_to_id.get(molecule, -1)
                    for molecule in molecules
                ),
            }

            product_nodes += [
                Node(
                    explore_weight=self.explore_weight,
                    scoring_fn=self.scoring_fn,
                    molecules=(product,),
                    unique_building_block_ids=node.unique_building_block_ids,
                    construction_log=node.construction_log + (reaction_log,),
                    rollout_num=self.rollout_num
                )
                for product in products
            ]

        return product_nodes

    def get_child_nodes(self, node: Node) -> list[Node]:
        """Get the child Nodes of a given Node.

        Child Nodes are created in two ways:
        1. By running all possible reactions on the molecules in the Node and creating Nodes with the products.
        2. By adding all possible next building blocks to the molecules in the Node and creating a new Node for each.

        :param node: The Node to get the child Nodes of.
        :return: A list of child Nodes of the given Node.
        """
        # Run all valid reactions on the current molecules to combine them into new molecules
        new_nodes = self.run_all_reactions(node=node)

        # Add all possible next building blocks to the current molecules in the Node
        if node.num_molecules == 0:
            next_building_blocks = self.all_building_blocks
        else:
            next_building_blocks = self.get_next_building_blocks(molecules=node.molecules)

        # Optionally, limit the number of next nodes
        if self.num_expand_nodes is not None and len(next_building_blocks) > self.num_expand_nodes:
            next_building_blocks = random_choice(
                rng=self.rng,
                array=next_building_blocks,
                size=self.num_expand_nodes,
                replace=False
            )

        # Convert next node molecule tuples into Node objects
        new_nodes += [
            Node(
                explore_weight=self.explore_weight,
                scoring_fn=self.scoring_fn,
                molecules=node.molecules + (next_building_block,),
                unique_building_block_ids=node.unique_building_block_ids | {next_building_block},
                construction_log=node.construction_log,
                rollout_num=self.rollout_num
            )
            for next_building_block in next_building_blocks
        ]

        # Remove duplicates but maintain order for reproducibility
        new_nodes = list(dict.fromkeys(new_nodes))

        return new_nodes

    def compute_mcts_score(self, node: Node, total_visit_count: int) -> float:
        """Computes the MCTS score of a Node.

        :param node: A Node.
        :param total_visit_count: The total number of visits to all nodes at the same level as this Node.
        :return: The MCTS score of the Node.
        """
        # Compute initial MCTS score
        mcts_score = node.Q() + node.U(n=total_visit_count)

        # Optionally encourage building block diversity by reducing MCTS score based on building block usage
        if self.building_block_diversity and node.num_molecules > 0:
            # Determine the maximum number of times any building block in the molecule has been used
            max_building_block_count = max(
                self.building_block_counts[building_block_id]
                for building_block_id in node.unique_building_block_ids
            )

            # Reduce MCTS score based on maximum building block usage
            # Note: count - 1 is used because every building block appears once as its own node
            mcts_score /= np.exp((max_building_block_count - 1) / 100)

        return mcts_score

    def rollout(self, node: Node) -> float:
        """Performs a generation rollout.

        :param node: A Node representing the root of the generation.
        :return: The value (reward) of the rollout.
        """
        if self.verbose:
            print(f'Node {node.node_id} (rollout {self.rollout_num})')
            print(f'Molecules = {node.molecules}')
            print(f'Num molecules = {node.num_molecules}')
            print(f'Num unique building blocks = {len(node.unique_building_block_ids)}')
            print(f'Num reactions = {node.num_reactions}')
            print(f'Score = {node.P}')
            print()

        # Stop the search if we've reached the maximum number of reactions
        if node.num_reactions >= self.max_reactions:
            return node.P

        # If this node has already been visited and the children have been stored, get its children from the dictionary
        if node in self.node_to_children:
            child_nodes = self.node_to_children[node]

        # Otherwise, expand the node to get its children
        else:
            # Expand the node both by running reactions with the current molecules and adding new building blocks
            child_nodes = self.get_child_nodes(node=node)

            # Check the node map and merge with an existing node if available
            child_nodes = [self.node_map.get(new_node, new_node) for new_node in child_nodes]

            # Add nodes with complete molecules to the node map
            for child_node in child_nodes:
                if child_node.num_molecules == 1 and child_node not in self.node_map:
                    child_node.node_id = len(self.node_map)
                    self.node_map[child_node] = child_node
                    self.building_block_counts.update(child_node.unique_building_block_ids)

            # Save the number of children in order to maintain a total node count
            node.num_children = len(child_nodes)

            # If storing nodes, store the children
            if self.store_nodes:
                self.node_to_children[node] = child_nodes

        # If no new nodes were generated, return the current node's value
        if len(child_nodes) == 0:
            if node.num_molecules == 1:
                return node.P
            else:
                raise ValueError('Failed to expand a partially expanded node.')

        # Select a node based on the MCTS score
        total_visit_count = sum(child_node.N for child_node in child_nodes)
        selected_node = self.optimization_fn(
            child_nodes,
            key=partial(self.compute_mcts_score, total_visit_count=total_visit_count)
        )

        # Check the node map and merge with an existing node if available
        if selected_node in self.node_map:
            selected_node = self.node_map[selected_node]
        # Otherwise, assign node ID and add to node map
        else:
            selected_node.node_id = len(self.node_map)
            self.node_map[selected_node] = selected_node

        # Unroll the selected node
        v = self.rollout(node=selected_node)

        # Get max whole molecule (non-building block) score across rollouts as feedback
        if selected_node.num_molecules == 1 and node.num_reactions > 0:
            v = max(v, selected_node.P)

        # Update exploit score and visit count
        selected_node.W += v
        selected_node.N += 1

        return v

    def generate(self, n_rollout: int) -> list[Node]:
        """Generate molecules for the specified number of rollouts.

        NOTE: Only returns Nodes with exactly one molecule and at least one reaction.

        :param n_rollout: The number of rollouts to perform.
        :return: A list of Node objects sorted from best to worst score.
        """
        # Set up rollout bounds
        rollout_start = self.rollout_num + 1
        rollout_end = rollout_start + n_rollout

        print(f'Running rollouts {rollout_start} through {rollout_end - 1}...')

        # Run the generation algorithm for the specified number of rollouts
        for rollout_num in trange(rollout_start, rollout_end):
            self.rollout_num = rollout_num
            self.rollout(node=self.root)

        # Get all the Nodes representing fully constructed molecules that are not building blocks within these rollouts
        nodes = [
            node
            for _, node in self.node_map.items()
            if node.num_molecules == 1 and node.num_reactions > 0 and (rollout_start <= node.rollout_num < rollout_end)
        ]

        # Sort by score and break ties by using Node ID
        nodes = sorted(
            nodes,
            key=lambda node: (node.P, (1 if self.ascending_scores else -1) * node.node_id),
            reverse=not self.ascending_scores
        )

        return nodes

    @property
    def approx_num_nodes_searched(self) -> int:
        """Gets the approximate number of nodes seen during the search.

        Note: This will over count any node that appears as a child node of multiple parent nodes.
        """
        return 1 + sum(node.num_children for node in self.node_map)

    @property
    def num_nodes_searched(self) -> int:
        """Gets the precise number of nodes seen during the search. Only possible if store_nodes is True."""
        if not self.store_nodes:
            raise ValueError('Cannot get the precise number of nodes searched if store_nodes is False.'
                             'Use approx_num_nodes_searched instead.')

        # Get a set of all nodes and child nodes that have been visited
        visited_nodes = set()
        for node, children in self.node_to_children.items():
            visited_nodes.add(node)
            visited_nodes.update(children)

        return len(visited_nodes)



"""Utility functions for synthemol reactions."""
import pickle
from pathlib import Path

from synthemol.reactions.reaction import Reaction


def set_all_building_blocks(
        reactions: tuple[Reaction],
        building_blocks: set[str]
) -> None:
    """Sets the allowed building block SMILES for all Reactions in a list of Reactions.

    Note: Modifies Reactions in place.

    :param reactions: A tuple of Reactions whose allowed building block SMILES will be set.
    :param building_blocks: A set of allowed building block SMILES.
    """
    for reaction in reactions:
        for reactant in reaction.reactants:
            reactant.all_building_blocks = building_blocks


def load_and_set_allowed_reaction_building_blocks(
        reactions: tuple[Reaction],
        reaction_to_reactant_to_building_blocks_path: Path
) -> None:
    """Loads a mapping of allowed building blocks for each reaction and sets the allowed SMILES for each reaction.

    :param reactions: A tuple of Reactions whose allowed SMILES will be set.
    :param reaction_to_reactant_to_building_blocks_path: Path to a PKL file mapping from reaction ID
                                                         to reactant index to a set of allowed building block SMILES.
    """
    # Load allowed building blocks for each reaction
    with open(reaction_to_reactant_to_building_blocks_path, 'rb') as f:
        reaction_to_reactant_to_building_blocks: dict[int, dict[int, set[str]]] = pickle.load(f)

    # Set allowed building blocks for each reaction
    for reaction in reactions:
        for reactant_index, reactant in enumerate(reaction.reactants):
            reactant.allowed_building_blocks = reaction_to_reactant_to_building_blocks[reaction.id][reactant_index]



"""Contains training and predictions functions for scikit-learn models."""
import pickle
from pathlib import Path

import numpy as np
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor

from synthemol.constants import SKLEARN_MODEL_TYPES


def sklearn_load(
        model_path: Path
) -> SKLEARN_MODEL_TYPES:
    """Loads a scikit-learn model.

    :param model_path: Path to a scikit-learn model.
    :return: A scikit-learn model.
    """
    with open(model_path, 'rb') as f:
        model = pickle.load(f)

    return model


def sklearn_predict(
        model: SKLEARN_MODEL_TYPES,
        fingerprints: np.ndarray
) -> np.ndarray:
    """Predicts molecular properties using a scikit-learn model.

    :param model: A scikit-learn model.
    :param fingerprints: A 2D array of molecular fingerprints (num_molecules, num_features).
    :return: A 1D array of predicted properties (num_molecules,).
    """
    if isinstance(model, RandomForestClassifier) or isinstance(model, MLPClassifier):
        preds = model.predict_proba(fingerprints)[:, 1]
    elif isinstance(model, RandomForestRegressor) or isinstance(model, MLPRegressor):
        preds = model.predict(fingerprints)
    else:
        raise ValueError(f'Model type {type(model)} is not supported.')

    return preds


def sklearn_predict_on_molecule(
        model: SKLEARN_MODEL_TYPES,
        fingerprint: np.ndarray | None = None
) -> float:
    """Predicts the property of a molecule using a scikit-learn model.

    :param model: A scikit-learn model.
    :param fingerprint: A 1D array of molecular fingerprints (if applicable).
    :return: The model prediction on the molecule.
    """
    return sklearn_predict(model=model, fingerprints=fingerprint.reshape(1, -1))[0]


def sklearn_predict_on_molecule_ensemble(
        models: list[SKLEARN_MODEL_TYPES],
        fingerprint: np.ndarray | None = None
) -> float:
    """Predicts the property of a molecule using an ensemble of scikit-learn models.

    :param models: An ensemble of scikit-learn models.
    :param fingerprint: A 1D array of molecular fingerprints (if applicable).
    :return: The ensemble prediction on the molecule.
    """
    return float(np.mean([
        sklearn_predict_on_molecule(model=model, fingerprint=fingerprint)
        for model in models
    ]))



"""QueryMol class for storing a SMARTS query and associated helper functions."""
from functools import cache
from typing import Iterable

from rdkit import Chem

from synthemol.utils import convert_to_mol, strip_atom_mapping


class QueryMol:
    """Contains a molecule query in the form of a SMARTS string along with helper functions."""

    def __init__(self, smarts: str) -> None:
        """Initializes the QueryMol.

        :param smarts: A SMARTS string representing the molecular query.
        """
        self.smarts_with_atom_mapping = smarts
        self.smarts = strip_atom_mapping(smarts)
        self.query_mol = Chem.MolFromSmarts(self.smarts)

        # All building block SMILES
        self._all_building_block_set = self._all_building_block_list = None

        # Building block SMILES that are allowed to match this QueryMol
        self._allowed_building_block_set = self._allowed_building_block_list = None

    @property
    def all_building_blocks(self) -> list[str] | None:
        """Gets a set of all building block SMILES for this QueryMol."""
        return self._all_building_block_list

    @all_building_blocks.setter
    def all_building_blocks(self, all_building_blocks: Iterable[str]) -> None:
        """Sets the all building block SMILES for this QueryMol. Can only be set once.

        Note: Filters the all SMILES to only include those that match the SMARTS of this QueryMol.

        :param all_building_blocks: An iterable of building block SMILES that are allowed for this QueryMol.
        """
        if self._all_building_block_set is not None:
            raise ValueError('All building block SMILES has already been set.')

        self._all_building_block_set = set(all_building_blocks)
        self._all_building_block_list = sorted(self._all_building_block_set)

    @property
    def allowed_building_blocks(self) -> list[str] | None:
        """Gets a sorted list of allowed building block SMILES for this QueryMol."""
        return self._allowed_building_block_list

    @allowed_building_blocks.setter
    def allowed_building_blocks(self, allowed_building_blocks: Iterable[str]) -> None:
        """Sets the allowed building block SMILES for this QueryMol. Can only be set once.

        :param allowed_building_blocks: An iterable of building block SMILES that are allowed for this QueryMol.
        """
        if self._allowed_building_block_set is not None:
            raise ValueError('Allowed building block SMILES has already been set.')

        self._allowed_building_block_set = set(allowed_building_blocks)
        self._allowed_building_block_list = sorted(self._allowed_building_block_set)

    def has_substruct_match(self, smiles: str) -> bool:
        """Determines whether the provided molecule includes this QueryMol as a substructure.

        Note: Only checks SMARTS match, not self.allowed_set.

        :param smiles: A SMILES string
        :return: True if the molecule includes this QueryMol as a substructure, False otherwise.
        """
        mol = convert_to_mol(smiles, add_hs=True)

        return mol.HasSubstructMatch(self.query_mol)

    @cache
    def has_match(self, smiles: str) -> bool:
        """Determines whether the provided molecule matches this QueryMol.

        Always checks for SMARTS match and additionally checks for match in self.allowed_set
        if self.allowed_set is not None.

        Note: Caching assumes that self._allow_set does not change.

        :param smiles: A SMILES string.
        :return: True if the molecule matches this QueryMol, False otherwise.
        """
        # If SMILES is not in allowed building block set, return False (unless it's not a building block at all)
        if self._allowed_building_block_set is not None and smiles not in self._allowed_building_block_set:
            if self._all_building_block_set is not None and smiles not in self._all_building_block_set:
                pass
            else:
                return False

        # Check if SMILES matches SMARTS
        return self.has_substruct_match(smiles)

    def __str__(self) -> str:
        """Gets a string representation of the QueryMol."""
        return f'QueryMol({self.smarts})'



"""Contains training and predictions functions for Chemprop models."""
from pathlib import Path

import numpy as np
import torch
from chemprop.models import MoleculeModel
from chemprop.utils import load_checkpoint, load_scalers
from sklearn.preprocessing import StandardScaler


def chemprop_load(
        model_path: Path,
        device: torch.device = torch.device('cpu')
) -> MoleculeModel:
    """Loads a Chemprop model.

    :param model_path: A path to a Chemprop model.
    :param device: The device on which to load the model.
    :return: A Chemprop model.
    """
    return load_checkpoint(
        path=str(model_path),
        device=device
    ).eval()


def chemprop_load_scaler(
        model_path: Path
) -> StandardScaler:
    """Loads a Chemprop model's data scaler.

    :param model_path: A path to a Chemprop model.
    :return: A data scaler.
    """
    return load_scalers(path=str(model_path))[0]


def chemprop_predict_on_molecule(
        model: MoleculeModel,
        smiles: str,
        fingerprint: np.ndarray | None = None,
        scaler: StandardScaler | None = None
) -> float:
    """Predicts the property of a molecule using a Chemprop model.

    :param model: A Chemprop model.
    :param smiles: A SMILES string.
    :param fingerprint: A 1D array of molecular fingerprints (if applicable).
    :param scaler: A data scaler (if applicable).
    :return: The prediction on the molecule.
    """
    # Make prediction
    pred = model(
        batch=[[smiles]],
        features_batch=[fingerprint] if fingerprint is not None else None
    ).item()

    # Scale prediction if applicable
    if scaler is not None:
        pred = scaler.inverse_transform([[pred]])[0][0]

    return float(pred)


def chemprop_predict_on_molecule_ensemble(
        models: list[MoleculeModel],
        smiles: str,
        fingerprint: np.ndarray | None = None,
        scalers: list[StandardScaler] | None = None
) -> float:
    """Predicts the property of a molecule using an ensemble of Chemprop models.

    :param models: An ensemble of Chemprop models.
    :param smiles: A SMILES string.
    :param fingerprint: A 1D array of molecular fingerprints (if applicable).
    :param scalers: An ensemble of data scalers (if applicable).
    :return: The ensemble prediction on the molecule.
    """
    return float(np.mean([
        chemprop_predict_on_molecule(
            model=model,
            smiles=smiles,
            fingerprint=fingerprint,
            scaler=scaler
        ) for model, scaler in zip(models, scalers)
    ]))



"""Generate molecules combinatorially using a Monte Carlo tree search guided by a molecular property predictor."""
from datetime import datetime
from pathlib import Path

import pandas as pd
from tap import tapify

from synthemol.constants import (
    BUILDING_BLOCKS_PATH,
    FINGERPRINT_TYPES,
    MODEL_TYPES,
    OPTIMIZATION_TYPES,
    REACTION_TO_BUILDING_BLOCKS_PATH,
    REAL_BUILDING_BLOCK_ID_COL,
    SCORE_COL,
    SMILES_COL
)
from synthemol.reactions import (
    Reaction,
    REACTIONS,
    load_and_set_allowed_reaction_building_blocks,
    set_all_building_blocks
)
from synthemol.generate.generator import Generator
from synthemol.generate.utils import create_model_scoring_fn, save_generated_molecules


def generate(
        model_path: Path,
        model_type: MODEL_TYPES,
        save_dir: Path,
        building_blocks_path: Path = BUILDING_BLOCKS_PATH,
        fingerprint_type: FINGERPRINT_TYPES | None = None,
        reaction_to_building_blocks_path: Path | None = REACTION_TO_BUILDING_BLOCKS_PATH,
        building_blocks_id_column: str = REAL_BUILDING_BLOCK_ID_COL,
        building_blocks_score_column: str = SCORE_COL,
        building_blocks_smiles_column: str = SMILES_COL,
        reactions: tuple[Reaction] = REACTIONS,
        max_reactions: int = 1,
        n_rollout: int = 10,
        explore_weight: float = 10.0,
        num_expand_nodes: int | None = None,
        optimization: OPTIMIZATION_TYPES = 'maximize',
        rng_seed: int = 0,
        no_building_block_diversity: bool = False,
        store_nodes: bool = False,
        verbose: bool = False,
        replicate: bool = False
) -> None:
    """Generate molecules combinatorially using a Monte Carlo tree search guided by a molecular property predictor.

    :param model_path: Path to a directory of model checkpoints or to a specific PKL or PT file containing a trained model.
    :param model_type: Type of model to train.
    :param building_blocks_path: Path to CSV file containing molecular building blocks.
    :param save_dir: Path to directory where the generated molecules will be saved.
    :param fingerprint_type: Type of fingerprints to use as input features.
    :param reaction_to_building_blocks_path: Path to PKL file containing mapping from REAL reactions to allowed building blocks.
    :param building_blocks_id_column: Name of the column containing IDs for each building block.
    :param building_blocks_score_column: Name of column containing scores for each building block.
    :param building_blocks_smiles_column: Name of the column containing SMILES for each building block.
    :param reactions: A tuple of reactions that combine molecular building blocks.
    :param max_reactions: Maximum number of reactions that can be performed to expand building blocks into molecules.
    :param n_rollout: The number of times to run the generation process.
    :param explore_weight: The hyperparameter that encourages exploration.
    :param num_expand_nodes: The number of child nodes to include when expanding a given node. If None, all child nodes will be included.
    :param optimization: Whether to maximize or minimize the score.
    :param rng_seed: Seed for random number generators.
    :param no_building_block_diversity: Whether to turn off the score modification that encourages diverse building blocks.
    :param store_nodes: Whether to store in memory all the nodes of the search tree.
                        This doubles the speed of the search but significantly increases
                        the memory usage (e.g., 450 GB for 20,000 rollouts instead of 600 MB).
    :param replicate: This is necessary to replicate the results from the paper, but otherwise should not be used
                      since it limits the potential choices of building blocks.
    :param verbose: Whether to print out additional information during generation.
    """
    # Create save directory
    save_dir.mkdir(parents=True, exist_ok=True)

    # Load building blocks
    print('Loading building blocks...')

    # Optionally alter building blocks loading to precisely replicate previous experiments
    if replicate:
        # Change score loading dtype to ensure numerical precision
        building_block_data = pd.read_csv(building_blocks_path, dtype={building_blocks_score_column: str})
        building_block_data[building_blocks_score_column] = building_block_data[building_blocks_score_column].astype(float)

        # Reorder reactions
        old_reactions_order = [275592, 22, 11, 527, 2430, 2708, 240690, 2230, 2718, 40, 1458, 271948, 27]
        reactions = tuple(sorted(reactions, key=lambda reaction: old_reactions_order.index(reaction.id)))

        # Deduplicate building blocks by SMILES
        building_block_data.drop_duplicates(subset=building_blocks_smiles_column, inplace=True)
    # Otherwise, load the building blocks normally
    else:
        building_block_data = pd.read_csv(building_blocks_path)

    print(f'Loaded {len(building_block_data):,} building blocks')

    # Ensure unique building block IDs
    if building_block_data[building_blocks_id_column].nunique() != len(building_block_data):
        raise ValueError('Building block IDs are not unique.')

    # Map building blocks SMILES to IDs, IDs to SMILES, and SMILES to scores
    building_block_smiles_to_id = dict(zip(
        building_block_data[building_blocks_smiles_column],
        building_block_data[building_blocks_id_column]
    ))
    building_block_id_to_smiles = dict(zip(
        building_block_data[building_blocks_id_column],
        building_block_data[building_blocks_smiles_column]
    ))
    building_block_smiles_to_score = dict(zip(
        building_block_data[building_blocks_smiles_column],
        building_block_data[building_blocks_score_column]
    ))

    print(f'Found {len(building_block_smiles_to_id):,} unique building blocks')

    # Set all building blocks for each reaction
    set_all_building_blocks(
        reactions=reactions,
        building_blocks=set(building_block_smiles_to_id)
    )

    # Optionally, set allowed building blocks for each reaction
    if reaction_to_building_blocks_path is not None:
        print('Loading and setting allowed building blocks for each reaction...')
        load_and_set_allowed_reaction_building_blocks(
            reactions=reactions,
            reaction_to_reactant_to_building_blocks_path=reaction_to_building_blocks_path
        )

    # Define model scoring function
    print('Loading models and creating model scoring function...')
    model_scoring_fn = create_model_scoring_fn(
        model_path=model_path,
        model_type=model_type,
        fingerprint_type=fingerprint_type,
        smiles_to_score=building_block_smiles_to_score
    )

    # Set up Generator
    print('Setting up generator...')
    generator = Generator(
        building_block_smiles_to_id=building_block_smiles_to_id,
        max_reactions=max_reactions,
        scoring_fn=model_scoring_fn,
        explore_weight=explore_weight,
        num_expand_nodes=num_expand_nodes,
        optimization=optimization,
        reactions=reactions,
        rng_seed=rng_seed,
        no_building_block_diversity=no_building_block_diversity,
        store_nodes=store_nodes,
        verbose=verbose,
        replicate=replicate
    )

    # Search for molecules
    print('Generating molecules...')
    start_time = datetime.now()
    nodes = generator.generate(n_rollout=n_rollout)

    # Compute, print, and save stats
    stats = {
        'mcts_time': datetime.now() - start_time,
        'num_nonzero_reaction_molecules': len(nodes),
        'approx_num_nodes_searched': generator.approx_num_nodes_searched
    }

    print(f'MCTS time = {stats["mcts_time"]}')
    print(f'Number of full molecule, nonzero reaction nodes = {stats["num_nonzero_reaction_molecules"]:,}')
    print(f'Approximate total number of nodes searched = {stats["approx_num_nodes_searched"]:,}')

    if store_nodes:
        stats['num_nodes_searched'] = generator.num_nodes_searched
        print(f'Total number of nodes searched = {stats["num_nodes_searched"]:,}')

    pd.DataFrame(data=[stats]).to_csv(save_dir / 'mcts_stats.csv', index=False)

    # Save generated molecules
    print('Saving molecules...')
    save_generated_molecules(
        nodes=nodes,
        building_block_id_to_smiles=building_block_id_to_smiles,
        save_path=save_dir / 'molecules.csv'
    )


def generate_command_line() -> None:
    """Run generate function from command line."""
    tapify(generate)



"""Reaction class and helper functions."""
from rdkit import Chem
from rdkit.Chem import AllChem

from synthemol.reactions.query_mol import QueryMol
from synthemol.utils import convert_to_mol, MOLECULE_TYPE


class Reaction:
    """A chemical reaction including SMARTS for the reactants, product, and reaction along with helper functions."""

    def __init__(
            self,
            reactants: list[QueryMol],
            product: QueryMol,
            reaction_id: int | None = None
    ) -> None:
        """Initializes the Reaction.

        :param reactants: A list of QueryMols containing the reactants of the reaction.
        :param product: A QueryMol containing the product of the reaction.
        :param reaction_id: The ID of the reaction.
        """
        self.reactants = reactants
        self.product = product
        self.id = reaction_id

        self.reaction_smarts = f'{".".join(f"({reactant.smarts_with_atom_mapping})" for reactant in self.reactants)}' \
                               f'>>({self.product.smarts_with_atom_mapping})'
        self.reaction = AllChem.ReactionFromSmarts(self.reaction_smarts)

    def get_reactant_matches(self, smiles: str) -> list[int]:
        """Gets the indices of the reactants that match the provided SMILES.

        :param smiles: The SMILES to match.
        :return: A list of indices of the reactants that match the provided SMILES.
        """
        return [
            reactant_index
            for reactant_index, reactant in enumerate(self.reactants)
            if reactant.has_match(smiles)
        ]

    @property
    def num_reactants(self) -> int:
        """Gets the number of reactants in the reaction."""
        return len(self.reactants)

    def run_reactants(self, reactants: list[MOLECULE_TYPE]) -> tuple[tuple[Chem.Mol, ...], ...]:
        """Runs the reaction on the provided reactants.

        :param reactants: A list of reactants.
        :return: A tuple of tuples of RDKit Mol objects representing the products of the reaction.
        """
        return self.reaction.RunReactants([convert_to_mol(reactant, add_hs=True) for reactant in reactants])

    def __repr__(self) -> str:
        """Gets the string representation of the Reaction."""
        return f'{self.__class__.__name__}(id={self.id}, reaction={self.reaction_smarts})'



"""Utility functions for generating molecules."""
from functools import cache
from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
import torch
from chemfunc import compute_fingerprint

from synthemol.constants import FINGERPRINT_TYPES, MODEL_TYPES
from synthemol.generate.node import Node
from synthemol.models import (
    chemprop_load,
    chemprop_load_scaler,
    chemprop_predict_on_molecule_ensemble,
    sklearn_load,
    sklearn_predict_on_molecule_ensemble
)


def create_model_scoring_fn(
        model_path: Path,
        model_type: MODEL_TYPES,
        fingerprint_type: FINGERPRINT_TYPES | None = None,
        smiles_to_score: dict[str, float] | None = None
) -> Callable[[str], float]:
    """Creates a function that scores a molecule using a model or ensemble of models.

    :param model_path: A path to a model or directory of models.
    :param model_type: The type of model.
    :param fingerprint_type: The type of fingerprint to use.
    :param smiles_to_score: An optional dictionary mapping SMILES to precomputed scores.
    :return: A function that scores a molecule using a model or ensemble of models.
    """
    # Check compatibility of model and fingerprint type
    if model_type != 'chemprop' and fingerprint_type is None:
        raise ValueError('Must define fingerprint_type if using a scikit-learn model.')

    # Get model paths
    if model_path.is_dir():
        model_paths = list(model_path.glob('**/*.pt' if model_type == 'chemprop' else '**/*.pkl'))

        if len(model_paths) == 0:
            raise ValueError(f'Could not find any models in directory {model_path}.')
    else:
        model_paths = [model_path]

    # Load models and set up scoring function
    if model_type == 'chemprop':
        # Ensure reproducibility
        torch.manual_seed(0)
        torch.use_deterministic_algorithms(True)

        models = [chemprop_load(model_path=model_path) for model_path in model_paths]
        scalers = [chemprop_load_scaler(model_path=model_path) for model_path in model_paths]

        # Set up model scoring function for ensemble of chemprop models
        def model_scorer(smiles: str, fingerprint: np.ndarray | None = None) -> float:
            return chemprop_predict_on_molecule_ensemble(
                models=models,
                smiles=smiles,
                fingerprint=fingerprint,
                scalers=scalers
            )
    else:
        # Load scikit-learn models
        models = [sklearn_load(model_path=model_path) for model_path in model_paths]

        # Set up model scoring function for ensemble of scikit-learn models
        def model_scorer(smiles: str, fingerprint: np.ndarray) -> float:
            return sklearn_predict_on_molecule_ensemble(
                models=models,
                fingerprint=fingerprint
            )

    # Build model scoring function using either chemprop or scikit-learn ensemble and precomputed building block scores
    @cache
    def model_scoring_fn(smiles: str) -> float:
        if smiles_to_score is not None and smiles in smiles_to_score:
            return smiles_to_score[smiles]

        if fingerprint_type is not None:
            fingerprint = compute_fingerprint(smiles, fingerprint_type=fingerprint_type)
        else:
            fingerprint = None

        return model_scorer(
            smiles=smiles,
            fingerprint=fingerprint
        )

    return model_scoring_fn


def save_generated_molecules(
        nodes: list[Node],
        building_block_id_to_smiles: dict[int, str],
        save_path: Path
) -> None:
    """Save generated molecules to a CSV file.

    :param nodes: A list of Nodes containing molecules. Only nodes with a single molecule are saved.
    :param building_block_id_to_smiles: A dictionary mapping building block IDs to SMILES.
    :param save_path: A path to a CSV file where the molecules will be saved.
    """
    # Only keep nodes with one molecule
    nodes = [node for node in nodes if node.num_molecules == 1]

    # Convert construction logs from lists to dictionaries
    construction_dicts = []
    max_reaction_num = 0
    reaction_num_to_max_reactant_num = {}

    for node in nodes:
        construction_dict = {'num_reactions': len(node.construction_log)}
        max_reaction_num = max(max_reaction_num, len(node.construction_log))

        for reaction_index, reaction_log in enumerate(node.construction_log):
            reaction_num = reaction_index + 1
            construction_dict[f'reaction_{reaction_num}_id'] = reaction_log['reaction_id']

            reaction_num_to_max_reactant_num[reaction_num] = max(
                reaction_num_to_max_reactant_num.get(reaction_num, 0),
                len(reaction_log['building_block_ids'])
            )

            for reactant_index, building_block_id in enumerate(reaction_log['building_block_ids']):
                reactant_num = reactant_index + 1
                construction_dict[f'building_block_{reaction_num}_{reactant_num}_id'] = building_block_id
                construction_dict[f'building_block_{reaction_num}_{reactant_num}_smiles'] = building_block_id_to_smiles.get(building_block_id, '')

        construction_dicts.append(construction_dict)

    # Specify column order for CSV file
    columns = ['smiles', 'node_id', 'num_expansions', 'rollout_num', 'score', 'Q_value', 'num_reactions']

    for reaction_num in range(1, max_reaction_num + 1):
        columns.append(f'reaction_{reaction_num}_id')

        for reactant_num in range(1, reaction_num_to_max_reactant_num[reaction_num] + 1):
            columns.append(f'building_block_{reaction_num}_{reactant_num}_id')
            columns.append(f'building_block_{reaction_num}_{reactant_num}_smiles')

    # Save data
    save_path.parent.mkdir(parents=True, exist_ok=True)
    data = pd.DataFrame(
        data=[
            {
                'smiles': node.molecules[0],
                'node_id': node.node_id,
                'num_expansions': node.N,
                'rollout_num': node.rollout_num,
                'score': node.P,
                'Q_value': node.Q(),
                **construction_dict
            }
            for node, construction_dict in zip(nodes, construction_dicts)
        ],
        columns=columns
    )
    data.to_csv(save_path, index=False)



"""synthemol.generate module."""
from synthemol.generate.generate import generate
from synthemol.generate.generator import Generator
from synthemol.generate.node import Node
from synthemol.generate.utils import (
    create_model_scoring_fn,
    save_generated_molecules
)



"""Utility functions for synthemol."""
import re
from typing import Any

import numpy as np
from rdkit import Chem

from synthemol.constants import MOLECULE_TYPE


def strip_atom_mapping(
        smarts: str
) -> str:
    """Strips the atom mapping from a SMARTS (i.e., any ":" followed by digits).

    :param smarts: A SMARTS string with atom mapping indices.
    :return: The same SMARTS string but without the atom mapping indices.
    """
    return re.sub(r'\[([^:]+)(:\d+)]', r'[\1]', smarts)


def convert_to_mol(
        mol: MOLECULE_TYPE,
        add_hs: bool = False
) -> Chem.Mol:
    """Converts a SMILES to an RDKit Mol object (if not already converted) and optionally adds Hs.

    :param mol: A SMILES string or an RDKit Mol object.
    :param add_hs: Whether to add Hs.
    :return: An RDKit Mol object with Hs added optionally.
    """
    if isinstance(mol, str):
        mol = Chem.MolFromSmiles(mol)

    if add_hs:
        mol = Chem.AddHs(mol)

    return mol


def random_choice(
        rng: np.random.Generator,
        array: list[Any],
        size: int | None = None,
        replace: bool = True
) -> Any:
    """An efficient random choice function built on top of NumPy.

    :param rng: A NumPy random number generator.
    :param array: An array to choose from.
    :param size: The number of elements to choose. If None, a single element is chosen.
    :param replace: Whether to allow replacement.
    :return: A list of elements chosen from the array.
    """
    if size is None:
        return array[rng.integers(len(array))]

    return [
        array[i] for i in rng.choice(len(array), size=size, replace=replace)
    ]



"""Contains the Node class, which represents a step in the combinatorial molecule construction process."""
import math
from functools import cached_property
from typing import Any, Callable


class Node:
    """A Node represents a step in the combinatorial molecule construction process."""

    def __init__(
            self,
            explore_weight: float,
            scoring_fn: Callable[[str], float],
            node_id: int | None = None,
            molecules: tuple[str] | None = None,
            unique_building_block_ids: set[int] | None = None,
            construction_log: tuple[dict[str, Any]] | None = None,
            rollout_num: int | None = None
    ) -> None:
        """Initializes the Node.

        :param explore_weight: The hyperparameter that encourages exploration.
        :param scoring_fn: A function that takes as input a SMILES representing a molecule and returns a score.
        :param node_id: The ID of the Node, which should correspond to the order in which the Mode was visited.
        :param molecules: A tuple of SMILES. The first element is the currently constructed molecule
                          while the remaining elements are the building blocks that are about to be added.
        :param unique_building_block_ids: A set of building block IDS used in this Node.
        :param construction_log: A tuple of dictionaries containing information about each reaction
                                 used to construct the molecules in this Node.
        :param rollout_num: The number of the rollout on which this Node was created.
        """
        self.explore_weight = explore_weight
        self.scoring_fn = scoring_fn
        self.node_id = node_id
        self.molecules = molecules if molecules is not None else tuple()
        self.unique_building_block_ids = unique_building_block_ids if unique_building_block_ids is not None else set()
        self.construction_log = construction_log if construction_log is not None else tuple()
        self.W = 0.0  # The sum of the leaf Node values for leaf Nodes that descend from this Node.
        self.N = 0  # The number of times this Node has been expanded.
        self.rollout_num = rollout_num
        self.num_children = 0

    @classmethod
    def compute_score(cls, molecules: tuple[str], scoring_fn: Callable[[str], float]) -> float:
        """Computes the score of the molecules.

        :param molecules: A tuple of SMILES. The first element is the currently constructed molecule
                          while the remaining elements are the building blocks that are about to be added.
        :param scoring_fn: A function that takes as input a SMILES representing a molecule and returns a score.
        :return: The score of the molecules.
        """
        return sum(scoring_fn(molecule) for molecule in molecules) / len(molecules) if len(molecules) > 0 else 0.0

    @cached_property
    def P(self) -> float:
        """The property score of this Node. (Note: The value is cached, so it assumes the Node is immutable.)"""
        return self.compute_score(molecules=self.molecules, scoring_fn=self.scoring_fn)

    def Q(self) -> float:
        """Value that encourages exploitation of Nodes with high reward."""
        return self.W / self.N if self.N > 0 else 0.0

    def U(self, n: int) -> float:
        """Value that encourages exploration of Nodes with few visits."""
        return self.explore_weight * self.P * math.sqrt(1 + n) / (1 + self.N)

    @property
    def num_molecules(self) -> int:
        """Gets the number of building blocks in the Node."""
        return len(self.molecules)

    @property
    def num_reactions(self) -> int:
        """Gets the number of reactions used so far to generate the molecule in the Node."""
        return len(self.construction_log)

    def __hash__(self) -> int:
        """Hashes the Node based on the building blocks."""
        return hash(self.molecules)

    def __eq__(self, other: Any) -> bool:
        """Checks if the Node is equal to another Node based on the building blocks."""
        if not isinstance(other, Node):
            return False

        return self.molecules == other.molecules



"""synthemol.models module."""
from synthemol.models.chemprop_models import (
    chemprop_load,
    chemprop_load_scaler,
    chemprop_predict_on_molecule,
    chemprop_predict_on_molecule_ensemble
)
from synthemol.models.sklearn_models import (
    sklearn_load,
    sklearn_predict,
    sklearn_predict_on_molecule,
    sklearn_predict_on_molecule_ensemble
)



"""synthemol.reactions package."""
from synthemol.reactions.custom import CUSTOM_REACTIONS
from synthemol.reactions.query_mol import QueryMol
from synthemol.reactions.reaction import Reaction
from synthemol.reactions.real import REAL_REACTIONS
from synthemol.reactions.utils import (
    load_and_set_allowed_reaction_building_blocks,
    set_all_building_blocks
)

if CUSTOM_REACTIONS is None:
    REACTIONS: tuple[Reaction] = REAL_REACTIONS
else:
    REACTIONS: tuple[Reaction] = CUSTOM_REACTIONS



from pathlib import Path
from setuptools import find_packages, setup

# Load version number
__version__ = ''
version_file = Path(__file__).parent.absolute() / 'synthemol' / '_version.py'

with open(version_file) as fd:
    exec(fd.read())

# Load README
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='synthemol',
    version=__version__,
    author='Kyle Swanson',
    author_email='swansonk.14@gmail.com',
    description='synthemol',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/swansonk14/SyntheMol',
    download_url=f'https://github.com/swansonk14/SyntheMol/archive/refs/tags/v_{__version__}.tar.gz',
    license='MIT',
    packages=find_packages(),
    package_data={'synthemol': ['py.typed', 'resources/**/*']},
    entry_points={
        'console_scripts': [
            'synthemol=synthemol.generate.generate:generate_command_line'
        ]
    },
    install_requires=[
        'chemfunc',
        'chemprop',
        'descriptastorus',
        'matplotlib',
        'numpy',
        'pandas',
        'torch',
        'rdkit',
        'scikit-learn',
        'scipy',
        'tqdm',
        'typed-argument-parser>=1.8.0'
    ],
    python_requires='>=3.10',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Typing :: Typed'
    ],
    keywords=[
        'machine learning',
        'drug design',
        'generative models',
        'synthesizable molecules'
    ]
)
